// --------------------------------------------------------------------------------------------
#region // Copyright (c) 2003, SIL International. All Rights Reserved.   
// <copyright from='2003' to='2003' company='SIL International'>
//		Copyright (c) 2003, SIL International. All Rights Reserved.   
//    
//		Distributable under the terms of either the Common Public License or the
//		GNU Lesser General Public License, as specified in the LICENSING.txt file.
// </copyright> 
#endregion
// 
// File: Class1.cs
// Responsibility: Michael Paul Johnson
// Last reviewed: 
// 
// <remarks>
// </remarks>
// --------------------------------------------------------------------------------------------
using System;
using System.IO;
using System.Text;

namespace GBF
{
	public class GBFReader
	{
		protected StreamReader sr;
		protected string GBFFileName;
		protected string CurrentToken;
		protected enum TTextState 
		{
			Nothing, InTranslationTitle, InTranslationAbbr,
			InLongCopr, InShortCopr, InTitle, InShortTitle, InPsalmBookTitle, InHebrewTitle,
			InSectionTitle, InPlace, InContactName, InContactAddress, InHolyBibleTranslation,
			InCopyrightWork, InCopyrightYear, InCopyrightYearRange, InCopyrightOwner,
			InLanguageEthnologueCode, InLanguageName, InChapterSpecifier, InPsalmSpecifier,
			InVerseSpecifier, InRevisionComment
		};
		protected TTextState TextState;

		public enum TToken 
		{
			tokNull, tokEOF, tokHeader, tokContent, tokTail, tokStyle,
			tokString, tokSync, tokControl, tokFont};

		protected TToken TokenKind;
		public TToken GetTokenKind(){return TokenKind;}
		protected bool fEof;
		public bool Eof(){return fEof;}
		public string GBFVersion;


		public string sBook, sChapter, sVerse,
			sTitle,   // Title of this book of the Bible
			sShortTitle,
			sPsalmBookTitle, // Title of this Psalm book
			sHebrewTitle,    // Psalm Hebrew title
			sSectionTitle,   // Section headings
			sTranslationTitle,
			sTranslationAbbr,
			sLongCopr,
			sShortCopr,
			sPlace,
			sContactName,
			sContactAddressLine,
			sBookTitle,
			sCopyrightWork,
			sCopyrightYear,
			sCopyrightRange,
			sCopyrightOwner,
			sRevisionComment,
			sLanguageEthnologueCode,
			sLanguageName,
			sChapterSpecifier,
			sPsalmSpecifier,
			sVerseSpecifier;

		public string Token
		{
			get {return CurrentToken;}
		}

		public bool fPoetry;
		public int LineNumber;
		public int iBook, iChapter, iVerse;

		public void Open(string fileName)
		{
			string tok;

			sr = new StreamReader(fileName, System.Text.Encoding.Default);
			GBFFileName = fileName;
			fEof = false;
			TextState = TTextState.Nothing;
			LineNumber = 1;
			do	// Skip to file version header.
			{
				tok = GetToken();
			} while (!(tok.Length > 3) && (tok[0] == '<') && (tok[1] == 'H') && (tok[2] == '0'));
			GBFVersion = tok;
			iBook = iChapter = iVerse = 0;
			sTitle = sPsalmBookTitle = sHebrewTitle = sSectionTitle = "";
			sTranslationTitle = sTranslationAbbr = sLongCopr = sShortCopr = "";
			sPlace = sContactName = sContactAddressLine = "";
			sBookTitle = "Holy Bible";
			sCopyrightWork = sCopyrightYear = sCopyrightRange = sCopyrightOwner = "";
			sLanguageEthnologueCode = "";
			sLanguageName = "";
			sChapterSpecifier = "Chapter";
			sPsalmSpecifier = "Psalm";
			sVerseSpecifier = "Verse";
		}

		public GBFReader(string fileName)
		{
			this.Open(fileName);
		}

		protected char ReadCharacter()
		{
			if (fEof)
				return (char)0;
			int c = sr.Read();

			if (c == '\n')
				LineNumber++;
			if (c < 0)
			{
				fEof = true;
				sr.Close();
				return (char)0;
			}
			return (char) c;
		}

	protected string[] BookFileName =
	{
		"Preface","Genesis","Exodus","Lev","Num","Deut","Joshua","Judges", // 0 - 7
		"Ruth","1Sam","2Sam","1Kings","2Kings","1Chron","2Chron",   // 8 - 14
		"Ezra","Nehemiah","Esther","Job","Psalms","Proverbs",       // 15-20
		"Eccl","Song","Isaiah","Jeremiah","Lament","Ezekiel",       // 21-26
		"Daniel","Hosea","Joel","Amos","Obadiah","Jonah","Micah",   // 27-33
		"Nahum","Habakkuk","Zeph","Haggai","Zech","Malachi",        // 34-39
		"Tobit","Judith","GkEsther","Wisdom","Sirach","Baruch",       // 40-45
		"Let","Azar","Susanna","Bel","1Mac","2Mac","1Esdras",       // 46-52
		"Man","P1","3Mac","2Esdras","4Mac","AddDan","","","","","",       // 53-63
		"Matthew","Mark","Luke","John","Acts","Romans","1Cor",      // 64-70
		"2Cor","Gal","Eph","Philip","Col","1Thes","2Thes","1Tim",   // 71-78
		"2Tim","Titus","Philemon","Hebrews","James","1Peter",       // 79-84
		"2Peter","1John","2John","3John","Jude","Rev","helps"     // 85-90
	};

	public string[] osisBookName =
	{
		"Pref","Gen","Exod","Lev","Num","Deut","Josh","Judg", // 0 - 7
		"Ruth","1Sam","2Sam","1Kgs","2Kgs","1Chr","2Chr",   // 8 - 14
		"Ezra","Neh","Esth","Job","Ps","Prov",       // 15-20
		"Eccl","Song","Isa","Jer","Lam","Ezek",       // 21-26
		"Dan","Hos","Joel","Amos","Obad","Jonah","Mic",   // 27-33
		"Nah","Hab","Zeph","Hag","Zech","Mal",        // 34-39
		"Tob","Jdt","AddEsth","Wis","Sir","Bar",       // 40-45
		"EpJer","PrAzar","Sus","Bel","1Macc","2Macc","1Esd",       // 46-52
		"PrMan","Ps151","3Macc","2Esd","4Macc","AddDan","","","","","",       // 53-63
		"Matt","Mark","Luke","John","Acts","Rom","1Cor",      // 64-70
		"2Cor","Gal","Eph","Phil","Col","1Thess","2Thess","1Tim",   // 71-78
		"2Tim","Titus","Phlm","Heb","Jas","1Pet",       // 79-84
		"2Pet","1John","2John","3John","Jude","Rev","helps"     // 85-90
	};

	public string[] BookTLA =
	{
		"PRE","GEN","EXO","LEV","NUM","DEU","JOS","JDG", // 0 - 7
		"RUT","1SA","2SA","1KI","2KI","1CH","2CH",   // 8 - 14
		"EZR","NEH","EST","JOB","PSA","PRO",       // 15-20
		"ECC","SNG","ISA","JER","LAM","EZK",       // 21-26
		"DAN","HOS","JOL","AMO","OBA","JON","MIC",   // 27-33
		"NAM","HAB","ZEP","HAG","ZEC","MAL",        // 34-39
		"TOB","JDT","ESG","WIS","SIR","BAR",       // 40-45
		"LJE","S3Y","SUS","BEL","1MA","2MA","1ES",       // 46-52
		"MAN","PS2","3MA","2ES","4MA","DNT","","","","","",       // 53-63
		"MAT","MRK","LUK","JHN","ACT","ROM","1CO",      // 64-70
		"2CO","GAL","EPH","PHP","COL","1TH","2TH","1TI",   // 71-78
		"2TI","TIT","PHM","HEB","JAS","1PE",       // 79-84
		"2PE","1JN","2JN","3JN","JUD","REV",""     // 85-90
	};

		public string Reference()
		{
			return BookTLA[iBook]+" "+iChapter.ToString()+":"+iVerse.ToString();
		}

		public string OsisBook()
		{
			return osisBookName[iBook];
		}

		public string OsisChapterID()
		{
			return osisBookName[iBook]+"."+iChapter.ToString();
		}

		public string OsisVerseID()
		{
			return osisBookName[iBook]+"."+iChapter.ToString()+"."+iVerse.ToString();
		}

		public string Anchor()
		{
			return BookTLA[iBook]+iChapter.ToString()+"."+iVerse.ToString();
		}

		public string bookValue
		{
			get {return BookTLA[iBook];}
		}


		public string[] BookAbbr =
	{
		"","GE","EX","LV","NM","DT","JS","JG", // 0 - 7
		"RT","1S","2S","1K","2K","1CH","2CH",   // 8 - 14
		"EZR","NE","ES","JB","PS","PR",       // 15-20
		"EC","SS","IS","JE","LM","EZK",       // 21-26
		"DN","HS","JL","AM","OB","JO","MC",   // 27-33
		"NM","HB","ZP","HG","ZC","ML",        // 34-39
		"TOB","JDT","GR","WIS","SIR","BAR",       // 40-45
		"LJE","AZA","SUS","BEL","1MA","2MA","1ES",       // 46-52
		"MAN","PS2","3MA","2ES","4MA","","","","","","",       // 53-63
		"MT","MK","LK","JN","AC","RM","1CO",      // 64-70
		"2CO","GL","EP","PHP","CL","1TH","2TH","1TI",   // 71-78
		"2TI","TIT","PHM","HB","JS","1P",       // 79-84
		"2P","1J","2J","3J","JD","RV",""     // 85-90
	};

	

		protected string[]BookName =
{
	"Preface","Genesis","Exodus","Leviticus","Numbers","Deuteronomy","Joshua","Judges", // 0 - 7
	"Ruth","1 Samuel","2 Samuel","1 Kings","2 Kings","1 Chronicles","2 Chronicles",   // 8 - 14
	"Ezra","Nehemiah","Esther","Job","Psalms","Proverbs",       // 15-20
	"Ecclesiastes","Song of Songs","Isaiah","Jeremiah","Lamentations","Ezekiel",       // 21-26
	"Daniel","Hosea","Joel","Amos","Obadiah","Jonah","Micah",   // 27-33
	"Nahum","Habakkuk","Zephaniah","Haggai","Zechariah","Malachi",        // 34-39
	"Tobit","Judith","Esther (Greek)","Wisdom","Sirach","Baruch",       // 40-45
	"Letter of Jeremiah","Azariah","Susanna","Bel and the Dragon",
	"1 Maccabees","2 Maccabees","1 Esdras",       // 46-52
	"Prayer of Manasseh","Psalm 151","3 Maccabees","2 Esdras","4 Maccabees","Daniel (Greek)","","","","","",       // 53-63
	"Matthew","Mark","Luke","John","Acts","Romans","1 Corinthians",      // 64-70
	"2 Corinthians","Galatians","Ephesians","Philippians","Colossians",
	"1 Thessalonians","2 Thessalonians","1 Timothy",   // 71-78
	"2 Timothy","Titus","Philemon","Hebrews","James","1 Peter",       // 79-84
	"2 Peter","1 John","2 John","3 John","Jude","Revelation","helps"     // 85-90
};

		protected bool MatchAbbrev(string fullname, string abbrev)
		{
			int i, j;

			if (abbrev.Length > fullname.Length)
				return false;
			for (i = 0, j = 0; i < abbrev.Length; i++, j++)
			{
				if (abbrev[i] == ' ')
					i++;
				if (fullname[j] == ' ')
					j++;
				if (Char.ToUpper(abbrev[i]) != Char.ToUpper(fullname[j]))
					return false;
			}
			return true;
		}

		protected int BookNameToNumber(string abbr)
		{
			int i;

			for (i = 0; i < BookName.Length; i++)
			{
				if (MatchAbbrev(BookName[i], abbr))
					return i;
			}
			for (i = 0; i < BookFileName.Length; i++)
			{
				if (MatchAbbrev(BookFileName[i], abbr))
					return i;
			}
			for (i = 0; i < BookAbbr.Length; i++)
			{
				if (MatchAbbrev(BookAbbr[i], abbr))
					return i;
			}
			for (i = 0; i < BookTLA.Length; i++)
			{
				if (MatchAbbrev(BookTLA[i], abbr))
					return i;
			}
			return 0;
		}


		public string BookNumberToName(int BookNumber)
		{
			return BookName[BookNumber];
		}


		protected string peekedAt;

		public string GetToken()
		{
			char ch;

			if (peekedAt != null)
			{
				string result = peekedAt;
				peekedAt = null;
				return result;
			}

			if (fEof)
			{
				TokenKind = TToken.tokEOF;
				return "";
			}
			TokenKind = TToken.tokNull;
			StringBuilder Result = new StringBuilder("",1024);
			ch = ReadCharacter();
			Result.Append(ch);
			if (ch == '<')
			{	// Read to ">" or EOF
				do
				{
					ch = ReadCharacter();
					if (!fEof)
					{
						Result.Append(ch);
					}
				} while ((ch != '>') && (!fEof));
				switch (Result[1])
				{
					case 'B':
						TokenKind = TToken.tokContent;
						TextState = TTextState.Nothing;
						break;
					case 'C':
						TokenKind = TToken.tokControl;
						// Ignore all white space right after a paragraph end.
						// This makes the GBF source more readable, since a newline
						// can be inserted right after the <CM> or <CL> without
						// inserting any unwanted white space.
						while (Char.IsWhiteSpace((char)sr.Peek()))
						{
							ReadCharacter();
						}
						break;
					case 'D':
						TokenKind = TToken.tokControl;
						break;
					case 'H':
						TokenKind = TToken.tokHeader;
					switch (Result[2])
					{
						case '1':
							TextState = TTextState.InTranslationTitle;
							break;
						case '2':
							TextState = TTextState.InTranslationAbbr;
							break;
						case '3':
							TextState = TTextState.InShortCopr;
							break;
						case '4':
							TextState = TTextState.InLongCopr;
							break;
						case 'P':
							TextState = TTextState.InPlace;
							break;
						case 'C':
							TextState = TTextState.InContactName;
							break;
						case 'A':
							TextState = TTextState.InContactAddress;
							break;
						case 'T':
							TextState = TTextState.InHolyBibleTranslation;
							break;
						case 'W':
							TextState = TTextState.InCopyrightWork;
							break;
						case 'Y':
							TextState = TTextState.InCopyrightYear;
							break;
						case 'R':
							TextState = TTextState.InCopyrightYearRange;
							break;
						case 'O':
							TextState = TTextState.InCopyrightOwner;
							break;
						case 'E':
							TextState = TTextState.InLanguageEthnologueCode;
							break;
						case 'L':
							TextState = TTextState.InLanguageName;
							break;
						case 'S':
							TextState = TTextState.InChapterSpecifier;
							break;
						case 'U':
							TextState = TTextState.InPsalmSpecifier;
							break;
						case 'V':
							TextState = TTextState.InVerseSpecifier;
							break;
					}
						break;
					case 'F':
						TokenKind = TToken.tokFont;
						break;
					case 'J':
						TokenKind = TToken.tokStyle;
						break;
					case 'P':
						TokenKind = TToken.tokControl;
					switch (Result[2])
					{
						case 'P':
							fPoetry = true;
							break;
						case 'p':
							fPoetry = false;
							break;
					}
						break;
					case 'R':
						TokenKind = TToken.tokControl;
						break;
					case 'S':
					switch (Result[2])
					{
						case 'B':
							sBook = Result.ToString();
							sBook = sBook.Substring(3, sBook.Length - 4);
							sPsalmBookTitle = "";
							if (sBook == "")
							{
								iBook++;
								sBook = BookNumberToName(iBook);
							}
							else
							{
								iBook = BookNameToNumber(sBook);
							}
							sShortTitle = sBook;
							sTitle = sBook;
							iChapter = 0;
							iVerse = 0;
							break;
						case 'C':
							sChapter = Result.ToString();
							sChapter = sChapter.Substring(3, sChapter.Length - 4);
							if (sChapter == "")
								iChapter++;
							else
								iChapter = Int32.Parse(sChapter);
							iVerse = 0;
							break;
						case 'V':
							sVerse = Result.ToString();
							sVerse = sVerse.Substring(3, sVerse.Length - 4);
							if (sVerse == "")
								iVerse++;
							else
								iVerse = Int32.Parse(sVerse);
							break;
					}
						break;
					case 'T':
					switch (Result[2])
					{
						case 'B':
							TextState = TTextState.InPsalmBookTitle;
							sPsalmBookTitle = "";
							break;
						case 'b':
							TextState = TTextState.Nothing;
							break;
						case 'H':
							TextState = TTextState.InHebrewTitle;
							sHebrewTitle = "";
							break;
						case 'h':
							TextState = TTextState.Nothing;
							break;
						case 'S':
							TextState = TTextState.InSectionTitle;
							sSectionTitle = "";
							break;
						case 's':
							TextState = TTextState.Nothing;
							break;
						case 'T':
							TextState = TTextState.InTitle;
							sTitle = "";
							break;
						case 't':
							TextState = TTextState.Nothing;
							break;
						case 'N':
							TextState = TTextState.InShortTitle;
							sShortTitle = "";
							break;
						case 'n':
							TextState = TTextState.Nothing;
							sBook = sShortTitle;
							break;
						case 'C':
							TextState = TTextState.InRevisionComment;
							sRevisionComment = "";
							break;
						case 'c':
							TextState = TTextState.Nothing;
							break;
					}
						break;
					case 'Z':
						TextState = TTextState.Nothing;
						if (Result[2] == 'Z')
						{
							TokenKind = TToken.tokTail;
							fEof = true;
							sr.Close();
							iBook = 91;
						}
						break;
				}
			}
			else
			{	// Read everything up to but not including the next tag.
				while ((sr.Peek() != -1) && ('<' != (char)sr.Peek()))
				{
					Result.Append(ReadCharacter());
				}
				TokenKind = TToken.tokString;
			}
			CurrentToken = Result.ToString();
			if (TokenKind == TToken.tokString)
			{
				switch (TextState)
				{
					case TTextState.InHebrewTitle:
						sHebrewTitle = sHebrewTitle + CurrentToken;
						break;
					case TTextState.InLongCopr:
						sLongCopr = sLongCopr + CurrentToken;
						break;
					case TTextState.InPsalmBookTitle:
						sPsalmBookTitle += CurrentToken;
						break;
					case TTextState.InSectionTitle:
						sSectionTitle += CurrentToken;
						break;
					case TTextState.InShortCopr:
						sShortCopr += CurrentToken;
						break;
					case TTextState.InTitle:
						sTitle += CurrentToken;
						break;
					case TTextState.InShortTitle:
						sShortTitle += CurrentToken;
						break;
					case TTextState.InTranslationAbbr:
						sTranslationAbbr += CurrentToken;
						break;
					case TTextState.InTranslationTitle:
						sTranslationTitle += CurrentToken;
						break;
					case TTextState.InPlace:
                        sPlace = CurrentToken;
						break;
					case TTextState.InContactName:
						sContactName = CurrentToken;
						break;
					case TTextState.InContactAddress:
                        sContactAddressLine = CurrentToken;
						break;
					case TTextState.InHolyBibleTranslation:
                        sBookTitle = CurrentToken;
						break;
					case TTextState.InCopyrightWork:
                        sCopyrightWork = CurrentToken;
						break;
					case TTextState.InCopyrightYear:
						sCopyrightYear = CurrentToken;
						break;
					case TTextState.InCopyrightYearRange:
						sCopyrightRange = CurrentToken;
						break;
					case TTextState.InCopyrightOwner:
						sCopyrightOwner = CurrentToken;
						break;
					case TTextState.InLanguageEthnologueCode:
						sLanguageEthnologueCode = CurrentToken;
						break;
					case TTextState.InLanguageName:
						sLanguageName = CurrentToken;
						break;
					case TTextState.InChapterSpecifier:
						sChapterSpecifier = CurrentToken;
						break;
					case TTextState.InPsalmSpecifier:
						sPsalmSpecifier = CurrentToken;
						break;
					case TTextState.InVerseSpecifier:
						sVerseSpecifier = CurrentToken;
						break;
					case TTextState.InRevisionComment:
						sRevisionComment = sRevisionComment + CurrentToken;
						break;
				}
			}
			return CurrentToken;
		}

		public string PeekToken()
		{
			peekedAt = GetToken();
			return peekedAt;
		}
	}
}
