#region // Copyright (c) 2002-2003, SIL International. All Rights Reserved.   
// <copyright from='2002' to='2003' company='SIL International'>
//		Copyright (c) 2002-2003, SIL International. All Rights Reserved.   
//    
//		Distributable under the terms of either the Common Public License or the
//		GNU Lesser General Public License, as specified in the LICENSING.txt file.
// </copyright> 
#endregion
using System;
using System.IO;
using System.Text;
using System.Xml;
using GBF;

namespace GBF
{
	class OSISWriter
	{
		protected XmlTextWriter xml;
		protected GBFReader gbf;
		bool InDiv;
		bool InPoetry;
		bool InRed;
		bool IsItalic;
		bool LatentItalic;
		bool InVerse;
		bool InChapter;
		bool InLine;
		bool InLineGroup;
		bool InParagraph;
		bool InPsalmBookDiv;
		bool InPsalmTitleDiv;
		bool headOK;
		int QuotationLevel;
		string [] QuoteStart;
		const int maxQuoteLevel = 20;
		int ParagraphStarted;

		public string OsisDateNow()
		{
			return DateTime.UtcNow.ToString("yyyy-MM-dd");
		}

		public string OsisDateTimeNow()
		{
			return DateTime.UtcNow.ToString("yyyy-MM-ddTHH.mm.ss");
		}

		protected void BalanceQuotes(string s)
		{
			int i;
			int j = 0;


			for(i = 0; i < s.Length; i++)
			{
				if (j >= QuotationLevel)
					ParagraphStarted = 0;
				if (ParagraphStarted == 1)
					ParagraphStarted = 0;
				if (s[i] == '')
				{
					if (ParagraphStarted >= 2)
                    {
						j++;
						if ((j & 1) == 0)
						{
							Console.WriteLine("{0} quote level {1} paragraph reminder opens with ",
								gbf.Reference(), j);
						}
					}
					else
					{
						if (QuotationLevel < maxQuoteLevel)
							QuoteStart[QuotationLevel] = gbf.Reference();
						QuotationLevel++;
						if ((QuotationLevel & 1) == 0)
						{
							Console.WriteLine("{0} quote level {1} opens with ",
								gbf.Reference(), QuotationLevel);
							int k;
							for (k = QuotationLevel - 2; k >= 0; k--)
							{
								if (k < maxQuoteLevel)
                                    Console.WriteLine("    Level {0} started {1}", k+1, QuoteStart[k]);
							}
						}
					}
				}
				else if (s[i] == '')
				{
					if (ParagraphStarted >= 2)
					{
						if (j < QuotationLevel)
						{
							j++;
							if ((j & 1) == 1)
							{
								Console.WriteLine("{0} quote level {1} paragraph reminder opens with ",
									gbf.Reference(), j);
							}
						}
					}
					else
					{
						if (QuotationLevel < maxQuoteLevel)
                            QuoteStart[QuotationLevel] = gbf.Reference();
						QuotationLevel++;
						if ((QuotationLevel & 1) == 1)
						{
							Console.WriteLine("{0} quote level {1} opens with ",
								gbf.Reference(), QuotationLevel);
						}
					}
				}
				else if (s[i] == '')
				{
					if (QuotationLevel < 1)
					{
						Console.WriteLine("{0} unmatched closing quote ", gbf.Reference());
					}
					else
					{
						if ((QuotationLevel & 1) == 0)
						{
							Console.WriteLine("{0} quote level {1} closes with ", gbf.Reference(), QuotationLevel);
						}
						QuotationLevel--;
					}
				}
				else if (s[i] == '')
				{
					if ((s.Length <= (i+1)) || (!Char.IsLetter(s, i+1)))
					{
						if (((i > 0) && (s[i-1] != 's')) || (i == 0) ||
							((i > 3) && (s[i-3] == 'Y') && (s[i-2] == 'e') && (s[i-1] == 's')) ||
							((i > 4) && (s[i-4] == 'e') && (s[i-3] == 'y') && (s[i-2] == 'e') && (s[i-1] == 's')) ||
							((i > 4) && (s[i-4] == 'b') && (s[i-3] == 'e') && (s[i-2] == 'r') && (s[i-1] == 's')) ||
							((i > 4) && (s[i-4] == 's') && (s[i-3] == 'a') && (s[i-2] == 'y') && (s[i-1] == 's')) ||
							((i > 4) && (s[i-4] == 'g') && (s[i-3] == 'o') && (s[i-2] == 'd') && (s[i-1] == 's')))
						{
							if (!(((i > 4) && (s[i-4] == 'r') && (s[i-3] == 'i') && (s[i-2] == 'c') && (s[i-1] == 'e')) ||
								((i > 4) && (s[i-4] == 'e') && (s[i-3] == 'n') && (s[i-2] == 'c') && (s[i-1] == 'e')) ||
								((i > 4) && (s[i-4] == 'e') && (s[i-3] == 'l') && (s[i-2] == 'i') && (s[i-1] == 'x'))))
							{
								if (QuotationLevel < 1)
									Console.WriteLine("{0} unmatched closing quote ", gbf.Reference());
								else
								{
									if ((QuotationLevel & 1) == 1)
									{
										Console.WriteLine("{0} quote level {1} closes with ", gbf.Reference(), QuotationLevel);
									}
									QuotationLevel--;
								}
							}
						}
					}
				}
			
			}
			if (ParagraphStarted >= 2)
			{
				if (QuotationLevel != 0)
					Console.WriteLine("{0} missing expected open quote reminders.", gbf.Reference());
				ParagraphStarted = 0;
			}
		}

		protected void StartLine(int level)
		{
			headOK = false;
			if (InPoetry)
			{
				if (InParagraph)
				{
					xml.WriteEndElement();	// p
					InParagraph = false;
				}
				if (!InLineGroup)
				{
					xml.WriteStartElement("lg");
					InLineGroup = true;
					if (level == 0)
						level = 1;
				}
				else
				{
					if (level == 0)
						level = 2;
				}
				if (!InLine)
				{
					xml.WriteStartElement("l");
					if (level == 2)
						xml.WriteAttributeString("type", "x-secondary");
					else
						xml.WriteAttributeString("type", "x-primary");
					InLine = true;
				}
			}
			else
			{
				if (InLine)
				{
					xml.WriteEndElement();	// line
					InLine = false;
				}
				if (InLineGroup)
				{
					xml.WriteEndElement();	// lineGroup
					InLineGroup = false;
				}
				if (!InParagraph)
				{
					xml.WriteStartElement("p");
					InParagraph = true;
				}
			}
		}

		protected void EndLine()
		{
			if (IsItalic)
			{
				xml.WriteEndElement();	// transChange
				IsItalic = false;
				LatentItalic = true;
			}
			if (InLine)
			{
				xml.WriteEndElement();	// l
				InLine = false;
			}
			if (InLineGroup)
			{
				xml.WriteEndElement();	// lg
				InLineGroup = false;
			}
			if (InParagraph)
			{
				xml.WriteEndElement();	// p
				InParagraph = false;
			}
		}

		protected void StartItalic()
		{
			StartLine(0);
			xml.WriteStartElement("transChange");
			xml.WriteAttributeString("type", "added");
			IsItalic = true;
			LatentItalic = false;
		}

		protected void EndItalic()
		{
			if (!IsItalic && !LatentItalic)
			{
				Console.WriteLine("Redundant <Fi> at {0}", gbf.Reference());
			}
			LatentItalic = false;
			if (IsItalic)
			{
				xml.WriteEndElement();	// transChange
				IsItalic = false;
			}
		}

		protected string ReadFront()
		{
			string s;

			s = gbf.GetToken();
			while ((s != "") && !s.StartsWith("<B"))
			{
				if (s == "<FB>")
				{
					xml.WriteStartElement("title");
					xml.WriteAttributeString("type", "x-preface");
					xml.WriteString(gbf.GetToken());
					xml.WriteEndElement();	// title
				}
				else if (!s.StartsWith("<"))
				{
					xml.WriteElementString("p", s);
				}
				s = gbf.GetToken();
			}
			return s;
		}

		protected string ReadNote()
		{
			string s;

			xml.WriteStartElement("note");
			xml.WriteAttributeString("type", "translation");
			do
			{
				s = gbf.GetToken();
				if (s.StartsWith("<"))
				{
					if (s != "<Rf>")
						Console.WriteLine("Unexpected token {0} at {1}", s, gbf.Reference());
				}
				else
				{
					xml.WriteString(s);
				}
			} while (!gbf.Eof() && (s != "<Rf>"));
			xml.WriteEndElement();	// note
			return s;
		}

		protected string ReadBook(string StartToken)
		{
			int i;
			string s;
			bool startOfVerse = false;
			string verseList = "";
			string verse_sID = "";
			string chapterID = "";
			string verse_eID = "";
			string quoteID = "";
			int quoteNum = 0;
			
			headOK = false;
			s = StartToken;
			while (!s.StartsWith("<SB"))
			{
				Console.WriteLine("SKIPPED AT {0}: {1}", gbf.Reference(), s);
				s = gbf.GetToken();
			}
			xml.WriteStartElement("div");
			xml.WriteAttributeString("type", "book");
			xml.WriteAttributeString("osisID", gbf.OsisBook());
			xml.WriteAttributeString("scope", gbf.OsisBook());
			s = gbf.GetToken();	// Expecting <TT> or <TN>
			if (s == "<TN>")
			{
				do
				{
					s = gbf.GetToken();
					if (s.StartsWith("<") && (s != "<Tn>"))
						Console.WriteLine("UNEXPECTED TOKEN IN SHORT TITLE AT {0}: {1}", gbf.Reference(), s);
				} while ((s != "<Tn>") && !gbf.Eof());
				s = gbf.GetToken();	// Expecting <TT>
			}
			if (s == "<TT>")
			{
				xml.WriteStartElement("title");
				xml.WriteAttributeString("type", "main");
				xml.WriteAttributeString("short", gbf.sShortTitle);
				do
				{
					s = gbf.GetToken();
					if (!s.StartsWith("<"))
						xml.WriteString(s);
					else if (s == "<CM>")
						xml.WriteString(" ");
					else if (s != "<Tt>")
						Console.WriteLine("UNEXPECTED TOKEN IN TITLE AT {0}: {1}", gbf.Reference(), s);
				} while ((s != "<Tt>") && !gbf.Eof());
				xml.WriteEndElement();	// title
			}
			Console.Write("Reading {0}\r", gbf.bookValue);
			s = gbf.GetToken();
			while ((!gbf.Eof()) && (!s.StartsWith("<SB")) && (s != "<ZZ>")
				&& (s != "<BA>") && (s != "<BN>"))
			{
				if (s == "<WTP>")
				{
					xml.WriteStartElement("w");
					xml.WriteAttributeString("type", "x-plural");
					xml.WriteEndElement();
					s = gbf.GetToken();
				}
				if (s == "<WTf>")
				{
					xml.WriteStartElement("w");
					xml.WriteAttributeString("type", "x-directAddress");
					xml.WriteEndElement();
					s = gbf.GetToken();
				}
				if (startOfVerse && !s.StartsWith("<SV"))
				{
					StartLine(0);
					if (InVerse)
					{
						xml.WriteStartElement("verse");
						xml.WriteAttributeString("eID", verse_eID);
						xml.WriteEndElement();
						InVerse = false;
					}
					xml.WriteStartElement("verse");
					xml.WriteAttributeString("sID", verse_sID);
					verse_eID = verse_sID;
					xml.WriteAttributeString("osisID", verseList);
					xml.WriteEndElement();	// verse
					InVerse = true;
					verseList = "";
					startOfVerse = false;
				}
				if (IsItalic && s.StartsWith("<") && (s != "<Fi>") && (!s.StartsWith("<WT")))
				{
					xml.WriteEndElement();	// transChange
					IsItalic = false;
					LatentItalic = true;
				}
				if (s.StartsWith("<SC"))
				{
					if (InVerse)
					{
						xml.WriteStartElement("verse");
						xml.WriteAttributeString("eID", verse_eID);
						xml.WriteEndElement();
						InVerse = false;
					}
					if (InChapter)
					{
						xml.WriteStartElement("chapter");
						xml.WriteAttributeString("eID", chapterID);
						xml.WriteEndElement();
						InChapter = false;
					}
					xml.WriteStartElement("chapter");
					chapterID = gbf.OsisChapterID();
					xml.WriteAttributeString("sID", chapterID);
					xml.WriteAttributeString("osisID", chapterID);
					xml.WriteEndElement();	// chapter
					InChapter = true;
				}
				else if (s.StartsWith("<SV"))
				{
					if (verseList.Length == 0)
					{
						verseList = gbf.OsisVerseID();
						verse_sID = gbf.OsisVerseID();
					}
					else
					{
						verseList = verseList + " " + gbf.OsisVerseID();
					}
					startOfVerse = true;
				}
				else if (!s.StartsWith("<"))
				{
					StartLine(0);
					if (LatentItalic)
					{
						StartItalic();
					}
					xml.WriteString(s);
					BalanceQuotes(s);
				}
				else if (s == "<CM>")
				{
					EndLine();
					StartLine(1);
					if (InPoetry)
						ParagraphStarted++;
					else
						ParagraphStarted = 2;
				}
				else if (s == "<CL>")
				{
					if (InLine)
					{
						xml.WriteEndElement();	// l
						InLine = false;
					}
					if (InParagraph)
					{
						xml.WriteEndElement();	// p
						InParagraph = false;
					}
					StartLine(2);
					if (!InPoetry)
						ParagraphStarted++;
				}
				else if (s == "<RB>")
				{
					xml.WriteStartElement("milestone");
					xml.WriteAttributeString("type", "x-noteStartAnchor");
					xml.WriteEndElement();	// milestone
				}
				else if (s == "<RF>")
				{
					s = ReadNote();
				}
				else if (s == "<FR>")
				{
					if (InRed)
					{
						Console.WriteLine("Redundant <FR> at {0}", gbf.Reference());
					}
					else
					{
						InRed = true;
						xml.WriteStartElement("q");
						quoteNum++;
						quoteID = gbf.OsisVerseID()+"."+quoteNum.ToString();
						xml.WriteAttributeString("sID", quoteID);
						xml.WriteAttributeString("who", "Jesus");
						xml.WriteAttributeString("n", "");
						xml.WriteAttributeString("type", "x-doNotGeneratePunctuation");
						xml.WriteEndElement();	// q
					}
				}
				else if (s == "<Fr>")
				{
					if (!InRed)
					{
						Console.WriteLine("Redundant <Fr> at {0}", gbf.Reference());
					}
					else
					{
						xml.WriteStartElement("q");
						xml.WriteAttributeString("eID", quoteID);
						xml.WriteAttributeString("n", "");
						xml.WriteEndElement();
						InRed = false;
					}
				}
				else if (s == "<FI>")
				{
					StartItalic();
				}
				else if (s == "<Fi>")
				{
					EndItalic();
				}
				else if (s == "<PP>")
				{
					if (InPoetry)
					{
						Console.WriteLine("Redundant <PP> at {0}", gbf.Reference());
					}
					InPoetry = true;
				}
				else if (s == "<Pp>")
				{
					if (!InPoetry)
					{
						Console.WriteLine("Redundant <Pp> at {0}", gbf.Reference());
					}
					InPoetry = false;
				}
				else if (s == "<BP>")
				{
					do
					{
						s = gbf.GetToken();
						if (!s.StartsWith("<"))
						{
							StartLine(0);
							xml.WriteString(s);
						}
						else if (s == "<CM>")
						{
							EndLine();
						}
						else if (!s.StartsWith("<B"))
						{
							Console.WriteLine("{0} unhandled at {1}", s, gbf.Reference());
						}
					} while ((!gbf.Eof()) && (!s.StartsWith("<B")));
					if (InDiv)
					{
						xml.WriteEndElement();
						InDiv = false;
					}
				}
				else if (s == "<TH>")
				{
					StartLine(1);
				}
				else if (s != "<Th>")
				{
					EndLine();
				}
				else if (s == "<TS>")
				{
					EndLine();
					if (InDiv)
					{
						if (!headOK)
						{
							xml.WriteEndElement();	// div
							xml.WriteStartElement("div");
						}
					}
					else
					{
						xml.WriteStartElement("div");
						InDiv = true;
					}
					xml.WriteStartElement("title");
					xml.WriteAttributeString("type", "x-section");
					do
					{
						s = gbf.GetToken();
						if (!s.StartsWith("<"))
							xml.WriteString(s);
						else if (s == "<CM>")
							xml.WriteString(" ");
						else if (s != "<Ts>")
							Console.WriteLine("Unhandled token {0} at {1}", s, gbf.Reference());
					} while (!gbf.Eof() && (s != "<Ts>"));
					xml.WriteEndElement();	// title
					headOK = false;
				}
				else if (s == "<TB>")
				{
					if (InVerse)
					{
						xml.WriteStartElement("verse");
						xml.WriteAttributeString("eID", verse_eID);
						xml.WriteEndElement();
						InVerse = false;
					}
					if (InChapter)
					{
						xml.WriteStartElement("chapter");
						xml.WriteAttributeString("eID", chapterID);
						xml.WriteEndElement();
						InChapter = false;
					}
					EndLine();

					if (InPsalmTitleDiv)
					{
						xml.WriteEndElement();	// div
						InPsalmTitleDiv = false;
					}
					if (InPsalmBookDiv)
					{
						xml.WriteEndElement();	// div
						InPsalmBookDiv = false;
					}
					xml.WriteStartElement("div");
					InPsalmBookDiv = true;
					xml.WriteStartElement("title");
					xml.WriteAttributeString("type", "x-psalmBook");
					do
					{
						s = gbf.GetToken();
						if (!s.StartsWith("<"))
							xml.WriteString(s);
						else if (s == "<CM>")
							xml.WriteString(" ");
						else if (s != "<Tb>")
							Console.WriteLine("Unhandled tag {0} at {1}", s, gbf.Reference());
					} while (!gbf.Eof() && (s != "<Tb>"));
					xml.WriteEndElement();	// title
				}
				else if (s == "<TC>")
				{
					do	// Skip comment contents
					{
						s = gbf.GetToken();
					} while (!gbf.Eof() && (s != "<Tc>"));
				}
				else if (s == "<TP>")
				{
					StartLine(0);
					headOK = false;
				}
				else if (s == "<Tp>")
				{
					EndLine();
					if (InDiv)
					{
						xml.WriteEndElement();	// div
					}
					xml.WriteStartElement("div");
					InDiv = true;
					headOK = true;
				}
				s = gbf.GetToken();
			}


			if (InRed)
			{
				Console.WriteLine("Warning: dangling red at {0}", gbf.Reference());
				xml.WriteElementString("qEnd","");
				InRed = false;
			}
			if (IsItalic)
			{
				Console.WriteLine("Warning: dangling italic at {0}", gbf.Reference());
				EndItalic();
			}
			if (InVerse)
			{
				xml.WriteStartElement("verse");
				xml.WriteAttributeString("eID", verse_eID);
				xml.WriteEndElement();
				InVerse = false;
			}
			if (InChapter)
			{
				xml.WriteStartElement("chapter");
				xml.WriteAttributeString("eID", chapterID);
				xml.WriteEndElement();
				InChapter = false;
			}
			EndLine();
			if (InDiv)
			{
				xml.WriteEndElement();	// div
				InDiv = false;
			}
			if (InPsalmTitleDiv)
			{
				xml.WriteEndElement();	// div
				InPsalmTitleDiv = false;
			}
			if (InPsalmBookDiv)
			{
				xml.WriteEndElement();	// div
				InPsalmBookDiv = false;

			}
			xml.WriteEndElement();	// div type="book"
			if (QuotationLevel > 0)
			{
				for (i = QuotationLevel - 1; i >= 0; i--)
				{
					if (i < maxQuoteLevel)
					{
						Console.WriteLine("Unmatched quote starts at {0}", QuoteStart[i]);
					}
				}
			}
			QuotationLevel = 0;
			return (s);
		}

		public int ConvertBibleFormat(string gbfname, string osisname)
		{
			string s;
			string sLang;
			string sRefSystem = "Bible.KJV";
			InDiv = false;
			InPoetry = false;
			InRed = false;
			IsItalic = false;
			InVerse = false;
			InChapter = false;
			InLine = false;
			InLineGroup = false;
			InParagraph = false;
			InPsalmBookDiv = false;
			InPsalmTitleDiv = false;
			QuotationLevel = 0;
			QuoteStart = new string[maxQuoteLevel];
			ParagraphStarted = 0;

			gbf = new GBFReader(gbfname);
			Console.WriteLine("Reading {0}", gbfname);
			do
			{
				s = gbf.GetToken();
			} while (!s.StartsWith("<B"));
			xml = new XmlTextWriter(osisname, System.Text.Encoding.UTF8);
			xml.Formatting = Formatting.Indented;
			xml.Indentation = 1;
			xml.WriteStartDocument();
			xml.WriteStartElement("osis");
			xml.WriteAttributeString("xmlns", "http://www.bibletechnologies.net/2003/OSIS/namespace");
			xml.WriteAttributeString("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance");
			xml.WriteAttributeString("xsi:schemaLocation", "http://www.bibletechnologies.net/2003/OSIS/namespace osisCore.2.0.xsd");
			xml.WriteStartElement("osisText");
			xml.WriteAttributeString("osisIDWork", gbf.sTranslationAbbr);
			xml.WriteAttributeString("osisRefWork", "Bible");
			if (gbf.sLanguageEthnologueCode == "ENG")
			{
				sLang = "en";
				xml.WriteAttributeString("xml:lang", sLang);
			}
			else
			{
				sLang = gbf.sLanguageEthnologueCode;
				xml.WriteAttributeString("xml:lang", "x-SIL-"+gbf.sLanguageEthnologueCode);
			}
			xml.WriteStartElement("header");
			xml.WriteStartElement("revisionDesc");
			xml.WriteAttributeString("resp", gbf.sContactName);
			xml.WriteElementString("date", OsisDateTimeNow());
			if ((gbf.sRevisionComment != null) && (gbf.sRevisionComment.Length > 0))
				xml.WriteElementString("p", gbf.sRevisionComment);
			xml.WriteElementString("p","Converted "+Path.GetFileName(gbfname)+" in GBF to "+Path.GetFileName(osisname)+
				@" in
an XML format that is mostly compliant with OSIS 2.0 using gbf2osis.exe.
(Please see http://ebt.cx/translation/ for links to this software.)");
			xml.WriteElementString("p",
@"GBF and OSIS metadata fields do not exactly correspond to each other, so
the conversion is not perfect in the metadata. However, the Scripture portion
should be correct.");
			xml.WriteElementString("p",
@"No attempt was to convert quotation marks to structural markers using q or
speech elements, because this would require language and style-dependent
processing, and because the current OSIS specification is deficient in that
quotation mark processing is not guaranteed to produce the correct results
for all languages and translations. In English texts, the hard part of the
conversion to markup is figuring out what  means.
The other difficulty is that OSIS in no way guarantees that these punctuation
marks would be reconstituted properly by software that reads OSIS files
for anything other than modern English, and even then, it does not
accommodate all styles of punctuation and all cases.
We strongly recommend that anyone using OSIS NOT replace quotation mark
punctuation in any existing text with q or speech elements. It is better
for multiple language processing capabilities to leave the quotation
punctuation as part of the text. If you need the q or speech markup, then you
may supplement those punctuation marks with those markup elements, but specify
the n='' parameter in those elements to indicate that no generation of any
punctuation from those markup elements is required or desired. That way you
can have BOTH correct punctuation already in the text AND markup so that you
can automatically determine when you are in a quotation or not, independent
of language. This may be useful for a search by speaker, for example.");
			xml.WriteElementString("p",
@"The output of gbf2osis marks Jesus' words in a non-standard way using the q
element AND quotation marks if they were marked with FR/Fr markers in the GBF
file. The OSIS 2.0 specification requires that quotation marks be stripped out,
and reinserted by software that reads the OSIS files when q elements are used.
This is not acceptable for the reasons given above, and we choose not to do
that, but we used the q element with who='Jesus' to indicate Jesus' words.
Do not generate any additional punctuation due to these markers. The correct
punctuation is already in the text.");
			xml.WriteElementString("p",
@"OSIS does not currently support footnote start anchors. Therefore, these
start anchors have been represented with milestone elements, in case someone
might like to use them, for example, to start an href element in a conversion
to HTML. (OSIS sort of supports the same idea by allowing a catchword to be
defined within a footnote, but I did not implement the processing to convert
to this different way of doing things, and it isn't exactly the same, anyway.)");
			xml.WriteElementString("p",
@"Traditional psalm book titles are rendered as text rather than titles, because
the title element does not support containing transChange elements, as would be
required to encode the KJV text using OSIS title elements. This may actually be
a superior solution, anyway, in that the Masoretic text makes no such distinction
(even though many modern typeset Bibles do make a typographic distinction in this
case).");
			xml.WriteElementString("p",
@"The schema location headers were modified to use local copies rather than the
standard locations so that these files could be validated and used without an
Internet connection active at all times (very important for the developer's
remote island location), but you may wish to change them back.");
			xml.WriteEndElement();	// revisionDesc
			xml.WriteStartElement("work");
			xml.WriteAttributeString("osisWork", gbf.sTranslationAbbr);
			xml.WriteElementString("title", gbf.sTranslationTitle);
			xml.WriteElementString("creator", gbf.sTranslationAbbr+" committee");
			xml.WriteStartElement("date");
			xml.WriteAttributeString("event", "eversion");
			xml.WriteAttributeString("type", "Gregorian");
			xml.WriteString(OsisDateNow());
			xml.WriteEndElement();	// date
			xml.WriteElementString("publisher", gbf.sContactName);
			xml.WriteStartElement("type");
			xml.WriteAttributeString("type", "OSIS");
			xml.WriteString("Bible");
			xml.WriteEndElement();	// type
			xml.WriteStartElement("identifier");
			xml.WriteAttributeString("type", "OSIS");
			xml.WriteString("Bible."+sLang+"."+gbf.sTranslationAbbr);
			if ((gbf.sTranslationAbbr == "WEB") || (gbf.sTranslationAbbr == "HNV"))
			{
				xml.WriteString(".draft."+OsisDateNow());
				sRefSystem = "Bible.WEB";
			}
			xml.WriteEndElement();	// identifier
			xml.WriteElementString("source", "http://eBible.org/"+gbf.sTranslationAbbr.ToLower()+"/");
			xml.WriteStartElement("language");
			xml.WriteAttributeString("type", "SIL");
			xml.WriteString(gbf.sLanguageEthnologueCode);
			xml.WriteEndElement();	// language
			xml.WriteElementString("coverage", gbf.sPlace);
			xml.WriteElementString("rights", gbf.sLongCopr);

			// List books in GBF file in <scope> elements.

			int rangeStart = 0;
			int rangeEnd = 0;
			int previousBook = 0;

			while (!gbf.Eof())
			{
				// Read to next book starting point.
				while (!((s.Length > 3) && (s.Substring(0,3) == "<SB")) && (!gbf.Eof()))
				{
					s = gbf.GetToken();
				}
				s = gbf.GetToken();
				if ((gbf.iBook > 0) && (gbf.iBook < 91))
				{
					if (rangeStart == 0)
						rangeStart = rangeEnd = previousBook = gbf.iBook;
					else if (gbf.iBook == previousBook + 1)
					{	// Extend the range
						rangeEnd = previousBook = gbf.iBook;
						// End range at end of testament
						if ((rangeEnd == 39) || (rangeEnd == 63) || (rangeEnd == 90))
						{
							if (rangeStart != rangeEnd)
								xml.WriteElementString("scope", gbf.osisBookName[rangeStart]+"-"+gbf.osisBookName[rangeEnd]);
							else
								xml.WriteElementString("scope", gbf.osisBookName[rangeStart]);
							rangeStart = rangeEnd = previousBook = 0;
						}
					}
					else
					{	// Discontinuity: print out range & start new one.
						if (rangeStart != rangeEnd)
                     xml.WriteElementString("scope", gbf.osisBookName[rangeStart]+"-"+gbf.osisBookName[rangeEnd]);
						else
							xml.WriteElementString("scope", gbf.osisBookName[rangeStart]);
						rangeStart = rangeEnd = gbf.iBook;
						previousBook = rangeStart;
					}
				}
			}
			if (rangeStart != 0)
			{
				if (rangeEnd == 0)
					xml.WriteElementString("scope", gbf.osisBookName[rangeStart]);
				else
					xml.WriteElementString("scope", gbf.osisBookName[rangeStart]+"-"+gbf.osisBookName[rangeEnd]);
				rangeStart = rangeEnd = previousBook = 0;
			}
			xml.WriteElementString("refSystem", sRefSystem);
			xml.WriteEndElement();	// work
			xml.WriteEndElement();	// header
			Console.WriteLine("{0} ({1})", gbf.sTranslationTitle,	gbf.sTranslationAbbr);

			gbf.Open(gbfname);

			do	// Skip to Introduction, New Testament, Apocrypha, or New Testament
			{
				s = gbf.GetToken();
			} while (!s.StartsWith("<B"));
			while (!gbf.Eof())
			{
				switch (s)
				{
					case "<BI>":
						xml.WriteStartElement("div");
						xml.WriteAttributeString("type", "front");
						s = ReadFront();
						xml.WriteEndElement();	// div
						break;
					case "<BO>":
						xml.WriteStartElement("div");
						xml.WriteAttributeString("type", "bookGroup");
						xml.WriteAttributeString("canonical", "true");
						s = gbf.GetToken();
						do
						{
							s = ReadBook(s);
						} while (s.StartsWith("<SB"));
						xml.WriteEndElement();	// div
						break;
					case "<BA>":
						xml.WriteStartElement("div");
						xml.WriteAttributeString("type", "bookGroup");
						s = gbf.GetToken();
						do
						{
							s = ReadBook(s);
						} while (s.StartsWith("<SB"));
						xml.WriteEndElement();	// div
						break;
					case "<BN>":
						xml.WriteStartElement("div");
						xml.WriteAttributeString("type", "bookGroup");
						xml.WriteAttributeString("canonical", "true");
						s = gbf.GetToken();
						do
						{
							s = ReadBook(s);
						} while (s.StartsWith("<SB"));
						xml.WriteEndElement();	// div
						break;
					case "<ZZ>":
						s = gbf.GetToken();
						break;
					default:
						Console.WriteLine("UNEXPECTED AT {0}: {1}", gbf.Reference(), s);
						s = gbf.GetToken();
						break;
				}
			}
			
			xml.WriteEndElement();	// osisText
			xml.WriteEndElement();	// osis
			xml.WriteEndDocument();
			xml.Flush();
			xml.Close();
			Console.WriteLine("{0} written.", osisname);
			return 0;
		}

	}

	/// <summary>
	/// Summary description for Class1.
	/// </summary>
	class Class1
	{
		static void help()
		{
			Console.WriteLine("Syntax:");
			Console.WriteLine("GBF2OSIS gbfname osisname");
		}
		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static int Main(string[] args)
		{
			string gbfname = "";
			string osisname = "";

			Console.WriteLine("GBF2OSIS Copyright (C) 2004 SIL International.\n"+
				"Releasable under GNU LGPL or (at your option) the Common Public License.\n");
			foreach (string arg in args)
			{
				if (gbfname == "")
					gbfname = arg;
				else if (osisname == "")
					osisname = arg;
				else
				{
					help();
					return 1;
				}
			}
			if (osisname == "")
			{
				help();
				return 1;
			}
			OSISWriter osis = new OSISWriter();
			return osis.ConvertBibleFormat(gbfname, osisname);
		}
	}
}
