#region // Copyright (c) 2003-2004, SIL International. All Rights Reserved.   
// <copyright from='2003' to='2004' company='SIL International'>
//		Copyright (c) 2003, SIL International. All Rights Reserved.   
//    
//		Distributable under the terms of either the Common Public License or the
//		GNU Lesser General Public License, as specified in the LICENSING.txt file.
// </copyright> 
#endregion
using System;
using System.IO;
using System.Text;
using GBF;

namespace GBF
{
	
	class SFWriter
	{
		protected StreamWriter OutputFile;
		protected GBFReader gbf;
		bool InPoetry;
		bool InRed;
		bool IsItalic;
		bool LatentItalic;
		bool BOL;
		bool VirtualSpace;
		int NoteNumber;
		string SFBookName;
		string translationName;
		const int maxQuoteLevel = 20;

		protected void WriteEOL()
		{
			OutputFile.WriteLine("");
			BOL = true;
			VirtualSpace = false;
		}

		protected void WriteSFM(string marker, string content, bool firstcol)
		{
			if (firstcol && !BOL)
				OutputFile.WriteLine("");
			else if (VirtualSpace)
			{
				OutputFile.Write(" ");
			}
			OutputFile.Write("\\{0}", marker);
			if (content != "")
				OutputFile.Write(" {0}", content);
			if (!marker.EndsWith("*"))
				VirtualSpace = true;
			BOL = false;
		}

		protected void WriteSFM(string marker, string content)
		{
			WriteSFM(marker, content, true);
		}


		protected void WriteSFM(string marker)
		{
			WriteSFM(marker, "", true);
		}

		protected void WriteString(string s)
		{
			if (VirtualSpace)
			{
				OutputFile.Write(" ");
				VirtualSpace = false;
			}
			OutputFile.Write(s);
			BOL = s[s.Length-1] == '\n';
		}


		protected void NewLine()
		{
			if (InRed)
			{
				WriteSFM("wj*", "", false);
			}
			if (InPoetry)
			{
				WriteSFM("q2");
			}
			else
			{
				WriteSFM("p");
			}
			if (InRed)
			{
				WriteSFM("wj", "", false);
			}
		}

		protected void NewParagraph()
		{
			if (InRed)
			{
				WriteSFM("wj*",  "",  false);
			}
			if (InPoetry)
			{
				WriteSFM("q1");
			}
			else
			{
				WriteSFM("p");
				WriteEOL();
			}
			if (InRed)
			{
				WriteSFM("wj", "", false);
			}
		}

	
		protected void StartItalic()
		{
			WriteSFM("add", "", false);
			IsItalic = true;
			LatentItalic = false;
		}

		protected void EndItalic()
		{
			if (!IsItalic && !LatentItalic)
			{
				Console.WriteLine("Redundant <Fi> at {0}", gbf.Reference());
			}
			LatentItalic = false;
			if (IsItalic)
			{
				WriteSFM("add*", "", false);
				IsItalic = false;
			}
		}


		protected string ReadBook(string StartToken)
		{
			int i;
			int verseBridgeStart, verseBridgeEnd;
			bool StartOfVerse;
			string s;
			string peek;
			NoteNumber = 0;
			
			verseBridgeStart = 0;
			verseBridgeEnd = 0;
			StartOfVerse = false;

			BOL = true;
			VirtualSpace = false;
			s = StartToken;
			while (!s.StartsWith("<SB"))
			{
				Console.WriteLine("SKIPPED AT {0}: {1}", gbf.Reference(), s);
				s = gbf.GetToken();
			}
			i = gbf.iBook;
			if (i > 39)
			{
				if (i < 64)
				{
					SFBookName = String.Format("A{0:D2}-{1}-{2}.sfm", i-39, gbf.bookValue, translationName);
				}
				else
				{
					SFBookName = String.Format("{0:D2}-{1}-{2}.sfm", i - 24, gbf.bookValue, translationName);
				}
			}
			else
			{
				SFBookName = String.Format("{0:D2}-{1}-{2}.sfm", i, gbf.bookValue, translationName);
			}
			OutputFile = new StreamWriter(SFBookName, false, System.Text.Encoding.UTF8);
			DateTime dt = DateTime.Now;
			WriteSFM("id", gbf.bookValue+" "+SFBookName+" "+gbf.sTranslationTitle+" "+dt.ToLongDateString());
			WriteSFM("ide", "UTF-8");
			s = gbf.GetToken();	// Expecting <TT> or <TN>
			if (s == "<TN>")
			{
				WriteSFM("h");
				do
				{
					s = gbf.GetToken();
					if (!s.StartsWith("<"))
						WriteString(s);
					else if (s == "<CM>")
						WriteString(" ");
					else if (s != "<Tn>")
						Console.WriteLine("UNEXPECTED TOKEN IN SHORT TITLE AT {0}: {1}", gbf.Reference(), s);
				} while ((s != "<Tn>") && !gbf.Eof());
				s = gbf.GetToken();	// Expecting <TT>
			}
			if (s == "<TT>")
			{
				WriteSFM("mt");
				do
				{
					s = gbf.GetToken();
					if (!s.StartsWith("<"))
						WriteString(s);
					else if (s == "<CM>")
						WriteString(" ");
					else if (s != "<Tt>")
						Console.WriteLine("UNEXPECTED TOKEN IN TITLE AT {0}: {1}", gbf.Reference(), s);
				} while ((s != "<Tt>") && !gbf.Eof());
				WriteSFM("p");
			}
			Console.Write("Reading {0}\r", gbf.bookValue);
			s = gbf.GetToken();
			while ((!gbf.Eof()) && (!s.StartsWith("<SB")) && (s != "<ZZ>")
				&& (s != "<BA>") && (s != "<BN>"))
			{
				if (s == "<CM>")
				{
					peek = gbf.PeekToken();
					if ((peek == "<JR>") || (peek  == "<JL>"))
					{
						s = gbf.GetToken();
						peek = gbf.PeekToken();
					}
					while (peek == "<CM>")
					{
						WriteSFM("b");
						s = gbf.GetToken();
						peek = gbf.PeekToken();
					}
					NewParagraph();
				}
				if (s == "<CL>")
				{
					NewLine();
				}
				if (IsItalic && s.StartsWith("<") && (s != "<Fi>") && (!s.StartsWith("<WT")))
				{
					IsItalic = false;
					LatentItalic = true;
				}
				if (StartOfVerse && !s.StartsWith("<SV"))
				{
					if (verseBridgeEnd != 0)
						WriteSFM("v", verseBridgeStart.ToString()+"-"+verseBridgeEnd.ToString());
					else
						WriteSFM("v", verseBridgeStart.ToString());
					verseBridgeStart = verseBridgeEnd = 0;
					StartOfVerse = false;
				}
				if (s.StartsWith("<SC"))
				{
					WriteSFM("c");
					WriteString(gbf.iChapter.ToString());
					WriteEOL();
				}
				else if (s.StartsWith("<SV"))
				{
					StartOfVerse = true;
					if (verseBridgeStart == 0)
						verseBridgeStart = gbf.iVerse;
					else
						verseBridgeEnd = gbf.iVerse;
				}
				else if (!s.StartsWith("<"))
				{
					if (LatentItalic)
					{
						StartItalic();
					}
					WriteString(s);
				}
				else if (s == "<RB>")
				{
					// No provision for link start in SF.
				}
				else if (s == "<RF>")
				{
					NoteNumber++;
					WriteSFM("f", "+", false);
					do
					{
						s = gbf.GetToken();
						if (s.StartsWith("<"))
						{
							if (s != "<Rf>")
								Console.WriteLine("Unexpected token {0} at {1}", s, gbf.Reference());
						}
						else
						{
							WriteString(s);
						}
					} while (!gbf.Eof() && (s != "<Rf>"));
					WriteSFM("f*", "", false);
				}
				else if (s == "<FR>")
				{
					if (InRed)
					{
						Console.WriteLine("Redundant <FR> at {0}", gbf.Reference());
					}
					else
					{
						InRed = true;
						WriteSFM("wj", "", false);
					}
				}
				else if (s == "<Fr>")
				{
					if (!InRed)
					{
						Console.WriteLine("Redundant <Fr> at {0}", gbf.Reference());
					}
					else
					{
						InRed = false;
						WriteSFM("wj*", "", false);
					}
				}
				else if (s == "<FI>")
				{
					StartItalic();
				}
				else if (s == "<Fi>")
				{
					EndItalic();
				}
				else if (s == "<PP>")
				{
					if (InPoetry)
					{
						Console.WriteLine("Redundant <PP> at {0}", gbf.Reference());
					}
					InPoetry = true;
				}
				else if (s == "<Pp>")
				{
					if (!InPoetry)
					{
						Console.WriteLine("Redundant <Pp> at {0}", gbf.Reference());
					}
					InPoetry = false;
				}
					/*************
				else if (s == "<BP>")
				{
					do
					{
						s = gbf.GetToken();
						if (!s.StartsWith("<"))
						{
							WriteString(s);
						}
						else if (s == "<CM>")
						{
							NewParagraph();
						}
						else if (!s.StartsWith("<B"))
						{
							Console.WriteLine("{0} unhandled at {1}", s, gbf.Reference());
						}
					} while ((!gbf.Eof()) && (!s.StartsWith("<B")));
				}
				********************/
				else if (s == "<TH>")
				{
					WriteSFM("d");
					do
					{
						s = gbf.GetToken();
						if (!s.StartsWith("<"))
							WriteString(s);
						else if (s == "<CM>")
						{
							// Ignore <CM> in Psalm header.
						}
						else if (s == "<RB>")
						{
						}
						else if (s == "<RF>")
						{
							NoteNumber++;
							WriteSFM("f", "+", false);
							do
							{
								s = gbf.GetToken();
								if (s.StartsWith("<"))
								{
									if (s != "<Rf>")
										Console.WriteLine("Unexpected token {0} at {1}", s, gbf.Reference());
								}
								else
								{
									WriteString(s);
								}
							} while (!gbf.Eof() && (s != "<Rf>"));
							WriteSFM("f*", "", false);
						}
						else if (s == "<FI>")
							StartItalic();
						else if (s == "<Fi>")
							EndItalic();
						else if (s == "<PP>")
						{
							if (InPoetry)
								Console.WriteLine("Redundant <PP> at {0}", gbf.Reference());
							InPoetry = true;
						}
						else if (s == "<Pp>")
						{
							if (!InPoetry)
								Console.WriteLine("Redundant <Pp> at {0}", gbf.Reference());
							InPoetry = false;
						}
						else if (s != "<Th>")
							Console.WriteLine("Unhandled token {0} at {1}", s, gbf.Reference());
					} while (!gbf.Eof() && (s != "<Th>"));
				}
				else if (s == "<TS>")
				{
					if (gbf.iBook == 22)
						WriteSFM("sp");
					else
						WriteSFM("s");
					do
					{
						s = gbf.GetToken();
						if (!s.StartsWith("<"))
							WriteString(s);
						else if (s == "<CM>")
							WriteString(" ");
						else if (s != "<Ts>")
							Console.WriteLine("Unhandled token {0} at {1}", s, gbf.Reference());
					} while (!gbf.Eof() && (s != "<Ts>"));
				}
				else if (s == "<TB>")
				{
					WriteSFM("ms1");
					do
					{
						s = gbf.GetToken();
						if (!s.StartsWith("<"))
							WriteString(s);
						else if (s == "<CM>")
							WriteString(" ");
						else if (s != "<Tb>")
							Console.WriteLine("Unhandled tag {0} at {1}", s, gbf.Reference());
					} while (!gbf.Eof() && (s != "<Tb>"));
					// NewParagraph();
				}
				else if (s == "<TC>")
				{
					do	// Skip comment contents
					{
						s = gbf.GetToken();
					} while (!gbf.Eof() && (s != "<Tc>"));
				}
				s = gbf.GetToken();
			}

			if (InRed)
			{
				InRed = false;
			}
			if (IsItalic)
			{
				Console.WriteLine("Warning: dangling italic at {0}", gbf.Reference());
				EndItalic();
			}
			WriteEOL();
			OutputFile.Close();
			return (s);
		}

		public int ConvertBibleFormat(string gbfname, string TranslationName)
		{
			string s;
			string introName;
			InPoetry = false;
			InRed = false;
			IsItalic = false;
			NoteNumber = 0;

			BOL = true;
			VirtualSpace = false;
			translationName = TranslationName;
			gbf = new GBFReader(gbfname);
			Console.WriteLine("Reading {0}", gbfname);

			do	// Skip to Introduction, New Testament, Apocrypha, or New Testament
			{
				s = gbf.GetToken();
			} while (!s.StartsWith("<B"));
			if (s == "<BI>")
			{	// Introduction.
				introName = String.Format("Intro-{0}.sfm", translationName);
				OutputFile = new StreamWriter(introName, false, System.Text.Encoding.UTF8);
				WriteSFM("id1", "Introduction");
				s = gbf.GetToken();
				do
				{
					if (s == "<FB>")
					{
						WriteSFM("p");
						WriteSFM("it");
						s = gbf.GetToken();
						if (s[0] != '<')
						{
							WriteString(s);
							s = gbf.GetToken();
						}
						if (s != "<Fb>")
						{
							Console.WriteLine("{0} line {1}: unexpected tag in preface title \"{2}\"",
								gbfname, gbf.LineNumber, s);
						}
						WriteSFM("p");
						s = gbf.GetToken();
					}
					do
					{
						if (s[0] != '<')
						{	// Text
							WriteString(s);
						}
						if (s == "<CM>")
						{
							WriteSFM("p");
						}
						s = gbf.GetToken();
					}
					while (((s[0] != '<') || (s == "<CM>")) && (!gbf.Eof()));
				} while (!s.StartsWith("<B"));
				OutputFile.Close();
			}
			
			while (!gbf.Eof())
			{
				s = ReadBook(s);
			}
			
			Console.WriteLine("{0} written.", translationName);
			return 0;
		}
	}




	/// <summary>
	/// Summary description for Class1.
	/// </summary>
	class SFGen
	{
		static void help()
		{
			Console.WriteLine("Syntax:");
			Console.WriteLine("GBF2SF gbfname translationName");
		}
		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static int Main(string[] args)
		{
			string gbfname = "";
			string translationName = "";

			Console.WriteLine("GBF2SF Copyright (C) 2002 SIL International.\n"+
				"Releasable under GNU LGPL or (at your option) the Common Public License.\n");
			foreach (string arg in args)
			{
				if (gbfname == "")
					gbfname = arg;
				else if (translationName == "")
					translationName = arg;
				else
				{
					help();
					return 1;
				}
			}
			if (translationName == "")
			{
				help();
				return 1;
			}
			SFWriter xsem = new SFWriter();
			return xsem.ConvertBibleFormat(gbfname, translationName);
		}
	}
}
